/**************************************************************************
*  This file is part of the TAL project (Tiny Abstraction Layer)
*
*  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de).
*  All rights reserved.
*
*  Based on "uart_17xx_40xx.h" from NXP, therefore:
*  Copyright(C) NXP Semiconductors, 2014
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without 
*  modification, are permitted provided that the following conditions 
*  are met:
*  
*  1. Redistributions of source code must retain the above copyright 
*     notice, this list of conditions and the following disclaimer.
*  2. Redistributions in binary form must reproduce the above copyright
*     notice, this list of conditions and the following disclaimer in the 
*     documentation and/or other materials provided with the distribution.
*  3. Neither the name of the author nor the names of its contributors may 
*     be used to endorse or promote products derived from this software 
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
*  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
*  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
*  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
*  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
*  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
*  SUCH DAMAGE.
*
***************************************************************************
*  History:
*
*  09.06.2017  mifi  First Version.
**************************************************************************/

/*
 * @brief LPC17xx/40xx UART chip driver
 *
 * @note
 * Copyright(C) NXP Semiconductors, 2014
 * All rights reserved.
 *
 * @par
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * LPC products.  This software is supplied "AS IS" without any warranties of
 * any kind, and NXP Semiconductors and its licensor disclaim any and
 * all warranties, express or implied, including all implied warranties of
 * merchantability, fitness for a particular purpose and non-infringement of
 * intellectual property rights.  NXP Semiconductors assumes no responsibility
 * or liability for the use of the software, conveys no license or rights under any
 * patent, copyright, mask work right, or any other intellectual property rights in
 * or to any products. NXP Semiconductors reserves the right to make changes
 * in the software without notification. NXP Semiconductors also makes no
 * representation or warranty that such application will be suitable for the
 * specified use without further testing or modification.
 *
 * @par
 * Permission to use, copy, modify, and distribute this software and its
 * documentation is hereby granted, under NXP Semiconductors' and its
 * licensor's relevant copyrights in the software, without fee, provided that it
 * is used in conjunction with NXP Semiconductors microcontrollers.  This
 * copyright, permission, and disclaimer notice must appear in all copies of
 * this code.
 */

#if !defined(__UART_40XX_H__)
#define __UART_40XX_H__

/**************************************************************************
*  Includes
**************************************************************************/

/**************************************************************************
*  Global Definitions
**************************************************************************/

typedef struct
{
   union {
      volatile uint32_t RBR;   /* Rx Buffer */
      volatile uint32_t THR;   /* Tx Holding */
      volatile uint32_t DLL;   /* Divisor Latch Low */
   };
   
   union {
      volatile uint32_t DLM;   /* Divisor Latch High */
      volatile uint32_t IER;   /* Interrupt Enable */
   };
   
   union {
      volatile uint32_t IIR;   /* Interrupt Identity Register (when read) */
      volatile uint32_t FCR;   /* FIFO Control Register (when written) */
   };
   
   volatile uint32_t LCR;      /* Line Control Register */
   volatile uint32_t MCR;      /* Modem Control Register */
   volatile uint32_t LSR;      /* Line Status Register */
   volatile uint32_t MSR;      /* Modem Status Register */
   volatile uint32_t SCR;      /* Scratchpad Register */
   volatile uint32_t ACR;      /* Auto-baud Control Register */
   volatile uint32_t Reserve2;
   volatile uint32_t FDR;      /* Fractional Divider Register */
   volatile uint32_t Reserve3;
   volatile uint32_t TER;      /* Transmit Enable Register */
} UART_TypeDef;

/**
 * @brief Macro defines for UART Interrupt Enable Register
 */
#define UART_IER_RBRINT         (1 << 0)     /*!< RBR Interrupt enable */
#define UART_IER_THREINT        (1 << 1)     /*!< THR Interrupt enable */
#define UART_IER_RLSINT         (1 << 2)     /*!< RX line status interrupt enable */
#define UART_IER_MSINT          (1 << 3)     /*!< Modem status interrupt enable - valid for 11xx, 17xx/40xx UART1, 18xx/43xx UART1  only */
#define UART_IER_CTSINT         (1 << 7)     /*!< CTS signal transition interrupt enable - valid for 17xx/40xx UART1, 18xx/43xx UART1 only */
#define UART_IER_ABEOINT        (1 << 8)     /*!< Enables the end of auto-baud interrupt */
#define UART_IER_ABTOINT        (1 << 9)     /*!< Enables the auto-baud time-out interrupt */
#define UART_IER_BITMASK        (0x307)      /*!< UART interrupt enable register bit mask  - valid for 13xx, 17xx/40xx UART0/2/3, 18xx/43xx UART0/2/3 only*/
#define UART1_IER_BITMASK       (0x30F)      /*!< UART1 interrupt enable register bit mask - valid for 11xx only */
#define UART2_IER_BITMASK       (0x38F)      /*!< UART2 interrupt enable register bit mask - valid for 17xx/40xx UART1, 18xx/43xx UART1 only */

/**
 * @brief Macro defines for UART Interrupt Identification Register
 */
#define UART_IIR_INTSTAT_PEND   (1 << 0)     /*!< Interrupt pending status - Active low */
#define UART_IIR_FIFO_EN        (3 << 6)     /*!< These bits are equivalent to FCR[0] */
#define UART_IIR_ABEO_INT       (1 << 8)     /*!< End of auto-baud interrupt */
#define UART_IIR_ABTO_INT       (1 << 9)     /*!< Auto-baud time-out interrupt */
#define UART_IIR_BITMASK        (0x3CF)      /*!< UART interrupt identification register bit mask */

/* Interrupt ID bit definitions */
#define UART_IIR_INTID_MASK     (7 << 1)     /*!< Interrupt identification: Interrupt ID mask */
#define UART_IIR_INTID_RLS      (3 << 1)     /*!< Interrupt identification: Receive line interrupt */
#define UART_IIR_INTID_RDA      (2 << 1)     /*!< Interrupt identification: Receive data available interrupt */
#define UART_IIR_INTID_CTI      (6 << 1)     /*!< Interrupt identification: Character time-out indicator interrupt */
#define UART_IIR_INTID_THRE     (1 << 1)     /*!< Interrupt identification: THRE interrupt */
#define UART_IIR_INTID_MODEM    (0 << 1)     /*!< Interrupt identification: Modem interrupt */

/**
 * @brief Macro defines for UART FIFO Control Register
 */
#define UART_FCR_FIFO_EN        (1 << 0)     /*!< UART FIFO enable */
#define UART_FCR_RX_RS          (1 << 1)     /*!< UART RX FIFO reset */
#define UART_FCR_TX_RS          (1 << 2)     /*!< UART TX FIFO reset */
#define UART_FCR_DMAMODE_SEL    (1 << 3)     /*!< UART DMA mode selection - valid for 17xx/40xx, 18xx/43xx only */
#define UART_FCR_BITMASK        (0xCF)       /*!< UART FIFO control bit mask */

#define UART_TX_FIFO_SIZE       (16)

/* FIFO trigger level bit definitions */
#define UART_FCR_TRG_LEV0       (0)          /*!< UART FIFO trigger level 0: 1 character */
#define UART_FCR_TRG_LEV1       (1 << 6)     /*!< UART FIFO trigger level 1: 4 character */
#define UART_FCR_TRG_LEV2       (2 << 6)     /*!< UART FIFO trigger level 2: 8 character */
#define UART_FCR_TRG_LEV3       (3 << 6)     /*!< UART FIFO trigger level 3: 14 character */

/**
 * @brief Macro defines for UART Line Control Register
 */
/* UART word length select bit definitions */
#define UART_LCR_WLEN_MASK      (3 << 0)     /*!< UART word length select bit mask */
#define UART_LCR_WLEN5          (0 << 0)     /*!< UART word length select: 5 bit data mode */
#define UART_LCR_WLEN6          (1 << 0)     /*!< UART word length select: 6 bit data mode */
#define UART_LCR_WLEN7          (2 << 0)     /*!< UART word length select: 7 bit data mode */
#define UART_LCR_WLEN8          (3 << 0)     /*!< UART word length select: 8 bit data mode */

/* UART Stop bit select bit definitions */
#define UART_LCR_SBS_MASK       (1 << 2)     /*!< UART stop bit select: bit mask */
#define UART_LCR_SBS_1BIT       (0 << 2)     /*!< UART stop bit select: 1 stop bit */
#define UART_LCR_SBS_2BIT       (1 << 2)     /*!< UART stop bit select: 2 stop bits (in 5 bit data mode, 1.5 stop bits) */

/* UART Parity enable bit definitions */
#define UART_LCR_PARITY_EN      (1 << 3)     /*!< UART Parity Enable */
#define UART_LCR_PARITY_DIS     (0 << 3)     /*!< UART Parity Disable */
#define UART_LCR_PARITY_ODD     (0 << 4)     /*!< UART Parity select: Odd parity */
#define UART_LCR_PARITY_EVEN    (1 << 4)     /*!< UART Parity select: Even parity */
#define UART_LCR_PARITY_F_1     (2 << 4)     /*!< UART Parity select: Forced 1 stick parity */
#define UART_LCR_PARITY_F_0     (3 << 4)     /*!< UART Parity select: Forced 0 stick parity */
#define UART_LCR_BREAK_EN       (1 << 6)     /*!< UART Break transmission enable */
#define UART_LCR_DLAB_EN        (1 << 7)     /*!< UART Divisor Latches Access bit enable */
#define UART_LCR_BITMASK        (0xFF)       /*!< UART line control bit mask */

/**
 * @brief Macro defines for UART Modem Control Register
 */
#define UART_MCR_DTR_CTRL       (1 << 0)     /*!< Source for modem output pin DTR */
#define UART_MCR_RTS_CTRL       (1 << 1)     /*!< Source for modem output pin RTS */
#define UART_MCR_LOOPB_EN       (1 << 4)     /*!< Loop back mode select */
#define UART_MCR_AUTO_RTS_EN    (1 << 6)     /*!< Enable Auto RTS flow-control */
#define UART_MCR_AUTO_CTS_EN    (1 << 7)     /*!< Enable Auto CTS flow-control */
#define UART_MCR_BITMASK        (0xD3)       /*!< UART bit mask value */

/**
 * @brief Macro defines for UART Line Status Register
 */
#define UART_LSR_RDR            (1 << 0)     /*!< Line status: Receive data ready */
#define UART_LSR_OE             (1 << 1)     /*!< Line status: Overrun error */
#define UART_LSR_PE             (1 << 2)     /*!< Line status: Parity error */
#define UART_LSR_FE             (1 << 3)     /*!< Line status: Framing error */
#define UART_LSR_BI             (1 << 4)     /*!< Line status: Break interrupt */
#define UART_LSR_THRE           (1 << 5)     /*!< Line status: Transmit holding register empty */
#define UART_LSR_TEMT           (1 << 6)     /*!< Line status: Transmitter empty */
#define UART_LSR_RXFE           (1 << 7)     /*!< Line status: Error in RX FIFO */
#define UART_LSR_TXFE           (1 << 8)     /*!< Line status: Error in RX FIFO */
#define UART_LSR_BITMASK        (0xFF)       /*!< UART Line status bit mask */
#define UART1_LSR_BITMASK       (0x1FF)      /*!< UART1 Line status bit mask - valid for 11xx, 18xx/43xx UART0/2/3 only */

/**
 * @brief Macro defines for UART Modem Status Register
 */
#define UART_MSR_DELTA_CTS      (1 << 0)     /*!< Modem status: State change of input CTS */
#define UART_MSR_DELTA_DSR      (1 << 1)     /*!< Modem status: State change of input DSR */
#define UART_MSR_LO2HI_RI       (1 << 2)     /*!< Modem status: Low to high transition of input RI */
#define UART_MSR_DELTA_DCD      (1 << 3)     /*!< Modem status: State change of input DCD */
#define UART_MSR_CTS            (1 << 4)     /*!< Modem status: Clear To Send State */
#define UART_MSR_DSR            (1 << 5)     /*!< Modem status: Data Set Ready State */
#define UART_MSR_RI             (1 << 6)     /*!< Modem status: Ring Indicator State */
#define UART_MSR_DCD            (1 << 7)     /*!< Modem status: Data Carrier Detect State */
#define UART_MSR_BITMASK        (0xFF)       /*!< Modem status: MSR register bit-mask value */

/**
 * @brief Macro defines for UART Auto baudrate control register
 */
#define UART_ACR_START          (1 << 0)     /*!< UART Auto-baud start */
#define UART_ACR_MODE           (1 << 1)     /*!< UART Auto baudrate Mode 1 */
#define UART_ACR_AUTO_RESTART   (1 << 2)     /*!< UART Auto baudrate restart */
#define UART_ACR_ABEOINT_CLR    (1 << 8)     /*!< UART End of auto-baud interrupt clear */
#define UART_ACR_ABTOINT_CLR    (1 << 9)     /*!< UART Auto-baud time-out interrupt clear */
#define UART_ACR_BITMASK        (0x307)      /*!< UART Auto Baudrate register bit mask */

/**
 * Autobaud modes
 */
#define UART_ACR_MODE0          (0)          /*!< Auto baudrate Mode 0 */
#define UART_ACR_MODE1          (1)          /*!< Auto baudrate Mode 1 */


#endif /* !__UART_40XX_H__ */

/*** EOF ***/
